{************************************************************************
 * (C) Pavel Kubricht, 1997                                             *
 * Verze 1.0 z: unor, 1997                                              *
 ************************************************************************
 Obsah jednotky View602 : Implementace a vyvoz komponenty TView602

  Nevisualni komponenta TView602 zapouzdruje pohled do tabulky (do dotazu)
  vytvoreny ve WinBase602.
  Zapouzdreni se opira o Windows techniku zvanou subclassing.
  Jelikoz zname handle pohledu (poskytuje winbasovska funkce Open_view), muzeme
  definovat novou window proceduru (subclassing) a v ni zpracovat zpravy ci
  volat EventHandlery pro tyto zpravy.
 }

unit View602;

interface
uses SysUtils,Classes,WinBase,WinTypes,WinProcs,Messages,Forms;

type
TWBEvent = procedure(Sender:TObject; var Msg: TMessage) of object;

TView602 = class(TComponent)
private
  FWBviewdef  :array[0..100] of Char; // Specifikace pohledu, ktery ma byt otevren
  FWBbase     :tcursnum;              // Otevreny kursor nebo konstanta NO_REDIR
  FWBflags    :uns32;                 // pznaky
  FWBposition :LongInt;               // Cislo prvniho zaznamu
  FWBparent_id:window_id;             // Handle pohledu, kter m bt pod otevranm pohledem, nebo 0
  FWBhandle   :window_id;             // Handle otevreneho pohledu
// Predchozi parametry viz funkce Open_view v encyklopedii funkci

  FOldWndProc :TFarProc;        // Stara window procedura
  FInstance   :Pointer;         // Ukazatel na starou window proceduru
  FOnRecEnter:    TWBEvent;     // Udalosti vyvolane automatickymi zpravami
  FOnChange:      TWBEvent;     // z WinBase602 - vyznam zprav viz metoda NewWndProc
  FOnReset_Record:TWBEvent;
  FOnReset_View:  TWBEvent;
  FOnSubCursr:   TWBEvent;
  FOnSuperCursor: TWBEvent;
  FOnOtherWMCmd999:     TWBEvent; // Pro zpracovani zbylych automatickych zprav
  procedure SetWBviewdef(Value: string); //Pristupove metody pro property WBviewdef
  function  GetWBviewdef : string;
protected
  procedure NewWndProc(var Msg : TMessage); // Nova window procedura
public
  property WBparent_id:window_id read FWBparent_id write FWBparent_id;
  property WBHandle :window_id read FWBHandle;
 // Procedura Open otevre zadan pohled a definuje novou window proceduru:
  function Open:window_id; virtual;
  procedure Close; virtual;
  procedure Toggle; virtual;
  constructor Create(AOwner:TComponent); override;
  destructor Destroy;override;
published
  property WBviewdef:string read GetWBviewdef write SetWBviewdef;
  property WBbase :tcursnum read FWBbase write FWBbase default NO_REDIR;
  property WBflags:uns32 read FWBflags write FWBflags;
  property WBposition :LongInt read FWBposition write FWBposition;

  property OnRecEnter:TWBEvent read FOnRecEnter write FOnRecEnter;
  property OnChange: TWBEvent read FOnChange write FOnChange;
  property OnReset_Record:TWBEvent read FOnReset_Record write FOnReset_Record;
  property OnReset_View:TWBEvent read FOnReset_View write FOnReset_View ;
  property OnSubCursr:TWBEvent read FOnSubCursr write FOnSubCursr;
  property OnSuperCursor:TWBEvent read FOnSuperCursor write FOnSuperCursor;
  property OnOtherWMCmd999:TWBEvent read FOnOtherWMCmd999 write FOnOtherWMCmd999;
end;

procedure Register;

implementation

procedure TView602.SetWBviewdef(Value: string);
begin
  StrPCopy(FWBviewdef, Value);
end;

function TView602.GetWBviewdef : string;
begin
 Result := string(FWBviewdef);
end;

procedure TView602.NewWndProc(var Msg : TMessage);
{ Nova window procedura : Chcete-li vytvorit dalsi event handlery pro
winbasovsky pohled implementujicich odezvu na zpravu, muzete je volat v teto
metode (podobne jako zde jiz volane event handlery) }
begin
if (Msg.Msg = WM_COMMAND) and (LOWORD(Msg.wParam) = 999) then //automaticke zpravy
  case HIWORD(Msg.wParam) of
   NOTIF_RECENTER   :  { vybran novy zaznam v pohledu }
    if Assigned(FOnRecEnter) then FOnRecEnter(Self,Msg);
   NOTIF_CHANGE      :  { zmenena slozka pohledu }
    if Assigned(FOnChange) then FOnChange(Self,Msg);
   NOTIF_RESET_RECORD:  { vybrany zaznam se prekresluje }
    if Assigned(FOnReset_Record) then FOnReset_Record(Self,Msg);
   NOTIF_RESET_VIEW  :  { cely pohled se prekresluje }
    if Assigned(FOnReset_View) then FOnReset_View(Self,Msg);
   NOTIF_SUBCURSOR   :  { pohled prechazi k subkurzoru QBE }
    if Assigned(FOnSubCursr) then FOnSubCursr(Self,Msg);
   NOTIF_SUPERCURSOR :  { pohled se vraci k superkurzoru }
    if Assigned(FOnSuperCursor) then FOnSuperCursor(Self,Msg);
   else if Assigned(FOnOtherWMCmd999) then FOnOtherWMCmd999(Self,Msg);
  end;

// Volani stare window procedury pro defaultove zpracovani:
Msg.Result := CallWindowProc(FOldWndProc,FWBHandle,Msg.Msg,Msg.wParam,Msg.lParam);
end;

Function TView602.Open:window_id;
 const filter :fltproc = nil; // Misto filtrovaci fce (ktera ve verzi 5 neni pod-
 // porovana) lze pouzit window proceduru vytvorenou subclassingem ci spise
 // implementovat EventHandlery volane touto procedurou. Proto vzdy fltproc = nil
 begin
 Result := FWBhandle;
 if FWBHandle = 0 then // Otevrit jen jeden exemplar pohledu
 begin
  Open_view(FWBviewdef, FWBbase, FWBflags, FWBposition, filter, FWBparent_id, FWBhandle);
  Result := FWBhandle;
  { Subclassing = vytvoreni nove window procedury: }
  if FInstance <> nil then FreeObjectInstance(FInstance);
  // Ukazatel na starou window proceduru:
  FOldWndProc := pointer(GetWindowLong(FWBHandle,gwl_WndProc));
  FInstance := MakeObjectInstance(NewWndProc);//Ukazatel na novou window proceduru
  SetWindowLong(WBHandle,gwl_WndProc,Longint(FInstance));// Definovani nove
                                                         // window procedury
 end;
end;

procedure TView602.Close;
begin
 Close_view(FWBHandle);
 if FInstance <> nil then
 begin
  FreeObjectInstance(FInstance);
  FInstance := nil;
 end;
end;

procedure TView602.Toggle;
begin
 if FWBHandle = 0 then Open else close;
end;

constructor TView602.Create(AOwner:TComponent);
begin
 inherited Create(AOwner);
 FWBbase  := NO_REDIR;
end;

destructor TView602.Destroy;
begin
 if FInstance <> nil then
 begin
  FreeObjectInstance(FInstance);
  FInstance := nil;
 end;
 inherited Destroy;
end;

procedure Register;
begin
 RegisterComponents('WB602', [TView602]);
end;

end.
