{************************************************************************
 * (C) Pavel Kubricht, 1997                                             *
 * Verze 1.0 z: unor, 1997                                              *
 ************************************************************************
Obsah jednotky WBNavig: Implementace a vyvoz komponenty TWBNavig.

Komponenta TWBNavig
1) Nem propertisu DataSource, nebot nen urcena pro rozhran BDE, ale pro
winbsovsk API - napojeni TWBNavig na data je automatick - komponenta si
sama zjist prve aktivn pohled, pak tabulku (ci dotaz), nad kterou je tento
pohled vystaven a operace provd nad touto tabulkou (dotazem).
2) Je vybavena - oproti standardni komponente TDBNavigator - udalostmi typu
 OnBeforeDelete atd.
Potrebujeme-li tedy provest operaci spjatou s nekterm z tlacitek z
TWBNavig, umozuje komponenta TWBNavig jete pred provedenm tto operace
provst nejakou cinnost (napr. dotzat se uzivatele ci nejakou predbeznou
kontrolu dat) a prpadne - dle vsledku tto cinnosti - operaci zakzat.
S udalosti OnBeforeDelete souvisi nova propertisa WBIgnoreBefDel typu Boolean,
ktera po nastaveni na True zpusobi ignorovani udalosti OnBeforeDelete.
3) Nepodporuje tlactka Edit, Cansel a Post

******************************************************************************}

unit WBNavig;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls,Buttons,
  ExtCtrls,Dialogs ;

const
  InitRepeatPause = 400;  { prvni takt casovace (ms) }
  RepeatPause     = 100;  { takt rozjeteho casovace (ms)}

type
  TWBNavButton = class;// Individualni tlacitko navigatoru

  TNavGlyph = (ngEnabled, ngDisabled);
  TNavigateBtn = (nbFirst, nbPrior, nbNext, nbLast, nbInsert, nbDelete, nbRefresh);
  TButtonSet = set of TNavigateBtn;
  TWBNavButtonStyle = set of (nsAllowTimer, nsFocusRect);

  ENavClick = procedure (Sender: TObject; Button: TNavigateBtn) of object;
  // typ pro udalosti typu OnBeforexxxxxx:
  TWBNavEvent = procedure(Sender:TObject; var IsOk:Boolean) of object;

{ TWBNavig }

  TWBNavig = class(TCustomPanel)
  private
    FWBIgnoreBefDel:Boolean;
    FVisibleButtons: TButtonSet;
    FHints: TStrings;
    ButtonWidth: Integer;
    MinBtnSize: TPoint;
    FOnNavClick: ENavClick;
    FocusedButton: TNavigateBtn;
    FOnBPrior:  TWBNavEvent;
    FOnBNext:   TWBNavEvent;
    FOnBFirst:  TWBNavEvent;
    FOnBLast:   TWBNavEvent;
    FOnBInsert: TWBNavEvent;
    FOnBDelete: TWBNavEvent;
    FOnBRefresh:TWBNavEvent;
    procedure InitButtons;
    procedure InitHints;
    procedure Click(Sender: TObject);
    procedure BtnMouseDown (Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure SetVisible(Value: TButtonSet);
    procedure AdjustSize (var W: Integer; var H: Integer);
    procedure SetHints(Value: TStrings);
    procedure WMSize(var Message: TWMSize);  message WM_SIZE;
    procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
    procedure WMKillFocus(var Message: TWMKillFocus); message WM_KILLFOCUS;
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    // handler je spusten kdykoli dojde ke zmene propertisy Enabled
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure BtnVisiblePrior(Ano: Boolean);
    procedure BtnVisibleAll;
  protected
    Buttons: array[TNavigateBtn] of TWBNavButton;
    procedure ActiveChanged;
    procedure Loaded; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;

{$IFDEF DELPHI_2}
  procedure GetChildren(Proc: TGetChildProc); override;
{$ELSE}  //  Pro Delphi3
  procedure GetChildren(Proc: TGetChildProc; Root: TComponent); override;
{$ENDIF}

    procedure Prior; virtual;
    procedure Next;  virtual;
    procedure First; virtual;
    procedure Last;  virtual;
    procedure Insert;virtual;
    procedure Delete;virtual;
    procedure Refresh;virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
    procedure BtnClick(Index: TNavigateBtn);
  published
    property VisibleButtons: TButtonSet read FVisibleButtons write SetVisible
      default [nbFirst, nbPrior, nbNext, nbLast, nbInsert, nbDelete, nbRefresh];
    property Align;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Ctl3D;
    property Hints: TStrings read FHints write SetHints;
    property ParentCtl3D;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property WBIgnoreBefDel :Boolean read FWBIgnoreBefDel write FWBIgnoreBefDel;
    property OnBeforePrior: TWBNavEvent read FOnBPrior write FOnBPrior;
    property OnBeforeNext: TWBNavEvent read FOnBNext write FOnBNext;
    property OnBeforeFirst: TWBNavEvent read FOnBFirst write FOnBFirst;
    property OnBeforeLast: TWBNavEvent read FOnBLast write FOnBLast;
    property OnBeforeInsert: TWBNavEvent read FOnBInsert write FOnBInsert;
    property OnBeforeDelete: TWBNavEvent read FOnBDelete write FOnBDelete;
    property OnBeforeRefresh:TWBNavEvent read FOnBRefresh write FOnBRefresh;
    property OnClick: ENavClick read FOnNavClick write FOnNavClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnResize;
    property OnStartDrag;
  end;

{ TWBNavButton }

  TWBNavButton = class(TSpeedButton)
  private
    FIndex: TNavigateBtn;
    FNavStyle: TWBNavButtonStyle;
    FRepeatTimer: TTimer;
    procedure TimerExpired(Sender: TObject);
  protected
    procedure Paint; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
  public
    destructor Destroy; override;
    property NavStyle: TWBNavButtonStyle read FNavStyle write FNavStyle;
    property Index : TNavigateBtn read FIndex write FIndex;
  end;

procedure Register;

implementation
uses  DBCtrls, WinBase;

procedure Register;
begin
  RegisterComponents('WB602', [TWBNavig]);
end;

{ TWBNavig }

const
  SFirstRecord = 62532;
  SPriorRecord = 62533;
  SNextRecord = 62534;
  SLastRecord = 62535;
  SInsertRecord = 62536;
  SDeleteRecord = 62537;
  SEditRecord = 62538;
  SPostEdit = 62539;
  SCancelEdit = 62540;
  SRefreshRecord = 62541;

  BtnStateName: array[TNavGlyph] of PChar = ('EN', 'DI');
  BtnTypeName: array[TNavigateBtn] of PChar = ('FIRST', 'PRIOR', 'NEXT', 'LAST',
   'INSERT', 'DELETE', 'REFRESH');
  BtnHintId: array[TNavigateBtn] of Word = (SFirstRecord, SPriorRecord,
   SNextRecord, SLastRecord, SInsertRecord, SDeleteRecord, SRefreshRecord);

constructor TWBNavig.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle - [csAcceptsControls, csSetCaption] + [csOpaque];
  if not NewStyleControls then ControlStyle := ControlStyle + [csFramed];
  FVisibleButtons := [nbFirst, nbPrior, nbNext, nbLast, nbInsert,
    nbDelete, nbRefresh];
  FHints := TStringList.Create;
  InitButtons;
  BevelOuter := bvNone;
  BevelInner := bvNone;
  Width := 241;
  Height := 25;
  ButtonWidth := 0;
  FocusedButton := nbFirst;
end;

destructor TWBNavig.Destroy;
begin
  FHints.Free;
  inherited Destroy;
end;

procedure TWBNavig.InitButtons;
var
  I: TNavigateBtn;
  Btn: TWBNavButton;
  X: Integer;
  ResName: string;
begin
  MinBtnSize := Point(20, 18);
  X := 0;
  for I := Low(Buttons) to High(Buttons) do
  begin
    Btn := TWBNavButton.Create (Self);
    Btn.Index := I;
    Btn.Visible := I in FVisibleButtons;
    Btn.Enabled := True;
    Btn.SetBounds (X, 0, MinBtnSize.X, MinBtnSize.Y);
    FmtStr(ResName, 'dbn_%s', [BtnTypeName[I]]);
    Btn.Glyph.Handle := LoadBitmap(HInstance, PChar(ResName));
    Btn.NumGlyphs := 2;
    Btn.Enabled := False;  {!!! Force creation of speedbutton images !!!}
    Btn.Enabled := True;
    Btn.OnClick := Click;
    Btn.OnMouseDown := BtnMouseDown;
    Btn.Parent := Self;
    Buttons[I] := Btn;
    X := X + MinBtnSize.X;
  end;
  InitHints;
  Buttons[nbPrior].NavStyle := Buttons[nbPrior].NavStyle + [nsAllowTimer];
  Buttons[nbNext].NavStyle  := Buttons[nbNext].NavStyle + [nsAllowTimer];
end;

procedure TWBNavig.InitHints;
var
  I: Integer;
  J: TNavigateBtn;
begin
  for J := Low(Buttons) to High(Buttons) do
    Buttons[J].Hint := LoadStr (BtnHintId[J]);
  J := Low(Buttons);
  for I := 0 to (FHints.Count - 1) do
  begin
    if FHints.Strings[I] <> '' then Buttons[J].Hint := FHints.Strings[I];
    if J = High(Buttons) then Exit;
    Inc(J);
  end;
end;

procedure TWBNavig.SetHints(Value: TStrings);
begin
  FHints.Assign(Value);
  InitHints;
end;

{$IFDEF DELPHI_2}    //Pro Delphi2.x
  procedure TWBNavig.GetChildren(Proc: TGetChildProc);
{$ELSE}  //  Pro Delphi3
  procedure TWBNavig.GetChildren(Proc: TGetChildProc; Root: TComponent);
{$ENDIF}
begin
end;

procedure TWBNavig.SetVisible(Value: TButtonSet);
var
  I: TNavigateBtn;
  W, H: Integer;
begin
  W := Width;
  H := Height;
  FVisibleButtons := Value;
  for I := Low(Buttons) to High(Buttons) do
    Buttons[I].Visible := I in FVisibleButtons;
  AdjustSize (W, H);
  if (W <> Width) or (H <> Height) then
    inherited SetBounds (Left, Top, W, H);
  Invalidate;
end;

procedure TWBNavig.AdjustSize (var W: Integer; var H: Integer);
var
  Count: Integer;
  MinW: Integer;
  I: TNavigateBtn;
  Space, Temp, Remain: Integer;
  X: Integer;
begin
  if (csLoading in ComponentState) then Exit;
  if Buttons[nbFirst] = nil then Exit;

  Count := 0;
  for I := Low(Buttons) to High(Buttons) do
  begin
    if Buttons[I].Visible then
    begin
      Inc(Count);
    end;
  end;
  if Count = 0 then Inc(Count);

  MinW := Count * MinBtnSize.X;
  if W < MinW then W := MinW;
  if H < MinBtnSize.Y then H := MinBtnSize.Y;

  ButtonWidth := W div Count;
  Temp := Count * ButtonWidth;
  if Align = alNone then W := Temp;

  X := 0;
  Remain := W - Temp;
  Temp := Count div 2;
  for I := Low(Buttons) to High(Buttons) do
  begin
    if Buttons[I].Visible then
    begin
      Space := 0;
      if Remain <> 0 then
      begin
        Dec(Temp, Remain);
        if Temp < 0 then
        begin
          Inc(Temp, Count);
          Space := 1;
        end;
      end;
      Buttons[I].SetBounds(X, 0, ButtonWidth + Space, Height);
      Inc(X, ButtonWidth + Space);
    end
    else
      Buttons[I].SetBounds (Width + 1, 0, ButtonWidth, Height);
  end;
end;

procedure TWBNavig.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
var
  W, H: Integer;
begin
  W := AWidth;
  H := AHeight;
  AdjustSize (W, H);
  inherited SetBounds (ALeft, ATop, W, H);
end;

procedure TWBNavig.WMSize(var Message: TWMSize);
var
  W, H: Integer;
begin
  inherited;

  { check for minimum size }
  W := Width;
  H := Height;
  AdjustSize (W, H);
  if (W <> Width) or (H <> Height) then
    inherited SetBounds(Left, Top, W, H);
  Message.Result := 0;
end;

procedure TWBNavig.Click(Sender: TObject);
var OkClick :Boolean;
begin
  OkClick := true;         // Nastaven implicitn hodnoty.
  if not (csDesigning in ComponentState) then
  with TWBNavButton(Sender) do
  case Index of
   nbPrior:  if Assigned(FOnBPrior)  then FOnBPrior(Self,OkClick);
   nbNext:   if Assigned(FOnBNext)   then FOnBNext(Self,OkClick);
   nbFirst:  if Assigned(FOnBFirst)  then FOnBFirst(Self,OkClick);
   nbLast:   if Assigned(FOnBLast)   then FOnBLast(Self,OkClick);
   nbInsert: if Assigned(FOnBInsert) then FOnBInsert(Self,OkClick);
   nbDelete:
    begin
     if not WBIgnoreBefDel then
     if Assigned(FOnBDelete) then FOnBDelete(Self,OkClick);
    end;
   nbRefresh: if Assigned(FOnBRefresh) then FOnBRefresh(Self,OkClick);
  end;
  if OkClick then BtnClick(TWBNavButton(Sender).Index);
// Pokracovn - provedeni operace - jen v prpade, ze v prslunm
// EventHandleru jsme nastavili OkClick na true.
end;

procedure TWBNavig.BtnMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  OldFocus: TNavigateBtn;
begin
  OldFocus := FocusedButton;
  FocusedButton := TWBNavButton (Sender).Index;
  if TabStop and (GetFocus <> Handle) and CanFocus then
  begin
    SetFocus;
    if (GetFocus <> Handle) then
      Exit;
  end
  else if TabStop and (GetFocus = Handle) and (OldFocus <> FocusedButton) then
  begin
    Buttons[OldFocus].Invalidate;
    Buttons[FocusedButton].Invalidate;
  end;
end;

procedure TWBNavig.BtnClick(Index: TNavigateBtn);
begin
 case Index of
   nbPrior:  Prior;
   nbNext :  Next;
   nbFirst:  First;
   nbLast :  Last;
   nbInsert: Insert;
   nbDelete: Delete;
   nbRefresh:Refresh;
 end;
 if not (csDesigning in ComponentState) and Assigned(FOnNavClick) then
   FOnNavClick(Self, Index);
end;

procedure TWBNavig.BtnVisibleAll;
begin
  Buttons[nbFirst].Enabled := true;
  Buttons[nbPrior].Enabled := true;
  Buttons[nbNext].Enabled  := true;
  Buttons[nbLast].Enabled  := true;
end;

procedure TWBNavig.BtnVisiblePrior(Ano: Boolean);
begin
  Buttons[nbFirst].Enabled := Ano;
  Buttons[nbPrior].Enabled := Ano;
  Buttons[nbNext].Enabled  := not Ano;
  Buttons[nbLast].Enabled  := not Ano;
end;

procedure TWBNavig.Prior;
var WinHandle : window_id;
    intcis, extcis: trecnum;
begin
 WinHandle := Active_view;
 Get_view_pos(WinHandle, intcis, extcis);
 if intcis > 0 then
 begin
  BtnVisibleAll;
//  dec(intcis);
//  Set_int_pos(WinHandle, intcis, -1);
  SendMessage(WinHandle, SZM_PREVREC, 0, 0);
 end else BtnVisiblePrior(False);
end;

procedure TWBNavig.Next;
var WinHandle : window_id;
    recnum, intcis, extcis, oldcis : trecnum;
    curs : tcursnum;
    fl:uns16;
begin
 WinHandle := Active_view;
 Get_fcursor(WinHandle, curs, fl);
 Rec_cnt(curs, recnum);
 Get_view_pos(WinHandle, intcis, extcis);
 oldcis := intcis;
//  inc(intcis);
//  Set_int_pos(WinHandle, intcis, -1);
 SendMessage(WinHandle, SZM_NEXTREC, 0, 0);
 Get_view_pos(WinHandle, intcis, extcis);
   if oldcis = intcis then
       BtnVisiblePrior(True)
   else BtnVisibleAll;

end;

procedure TWBNavig.First;
var WinHandle : window_id;
begin
 WinHandle := Active_view;
 SendMessage(WinHandle, SZM_FIRSTREC, 0, 0);
// Set_int_pos(WinHandle, 0, -1);
 BtnVisiblePrior(False);
end;

procedure TWBNavig.Last;
var WinHandle : window_id;
    curs : tcursnum;
    fl:integer;
    recnum: trecnum;
begin
 WinHandle := Active_view;
// Get_fcursor(WinHandle, curs, fl);
// Rec_cnt(curs, recnum);
// Set_int_pos(WinHandle, recnum - 1, -1);
 SendMessage(WinHandle, SZM_LASTREC, 0, 0);
 BtnVisiblePrior(True);
end;

procedure TWBNavig.Insert;
var WinHandle : window_id;
    curs : tcurstab;
    fl:uns16;
begin
 WinHandle := Active_view;
// SendMessage(WinHandle, SZM_INSERT, 0, 0);
// SendMessage(WinHandle, SZM_RESET, 0, 0);
 Get_fcursor(WinHandle, curs, fl);
 WinBase.Insert(curs);
 Reset_view(WinHandle,trecnum(-1),RESET_DELETIONS);
// procedure Reset_view(id : window_id;  rec : trecnum;  extent : uns16); stdcall;
end;

procedure TWBNavig.Delete;
var WinHandle : window_id;
    intcis, extcis : trecnum;
    curs : tcurstab;
    fl:uns16;
begin
 WinHandle := Active_view;
 Get_view_pos(WinHandle, intcis, extcis);
 Get_fcursor(WinHandle, curs, fl);
 WinBase.Delete(curs, extcis);
 Reset_view(WinHandle,trecnum(-1),RESET_DELETIONS);
end;

procedure TWBNavig.Refresh;
var WinHandle : window_id;
begin
  WinHandle := Active_view;
//  SendMessage(WinHandle, SZM_RESET, 0, 0);
  Reset_view(WinHandle,trecnum(-1),RESET_DELETIONS {+ RESET_CURSOR});
end;

procedure TWBNavig.WMSetFocus(var Message: TWMSetFocus);
begin
  Buttons[FocusedButton].Invalidate;
end;

procedure TWBNavig.WMKillFocus(var Message: TWMKillFocus);
begin
  Buttons[FocusedButton].Invalidate;
end;

procedure TWBNavig.KeyDown(var Key: Word; Shift: TShiftState);
var
  NewFocus: TNavigateBtn;
  OldFocus: TNavigateBtn;
begin
  OldFocus := FocusedButton;
  case Key of
    VK_RIGHT:
      begin
        NewFocus := FocusedButton;
        repeat
          if NewFocus < High(Buttons) then
            NewFocus := Succ(NewFocus);
        until (NewFocus = High(Buttons)) or (Buttons[NewFocus].Visible);
        if NewFocus <> FocusedButton then
        begin
          FocusedButton := NewFocus;
          Buttons[OldFocus].Invalidate;
          Buttons[FocusedButton].Invalidate;
        end;
      end;
    VK_LEFT:
      begin
        NewFocus := FocusedButton;
        repeat
          if NewFocus > Low(Buttons) then
            NewFocus := Pred(NewFocus);
        until (NewFocus = Low(Buttons)) or (Buttons[NewFocus].Visible);
        if NewFocus <> FocusedButton then
        begin
          FocusedButton := NewFocus;
          Buttons[OldFocus].Invalidate;
          Buttons[FocusedButton].Invalidate;
        end;
      end;
    VK_SPACE:
      begin
        if Buttons[FocusedButton].Enabled then
          Buttons[FocusedButton].Click;
      end;
  end;
end;

procedure TWBNavig.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  Message.Result := DLGC_WANTARROWS;
end;

procedure TWBNavig.ActiveChanged;
var
  I: TNavigateBtn;
begin
  if not Enabled then
    for I := Low(Buttons) to High(Buttons) do
      Buttons[I].Enabled := False
  else for I := Low(Buttons) to High(Buttons) do
      Buttons[I].Enabled := True;
end;

procedure TWBNavig.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  if not (csLoading in ComponentState) then
    ActiveChanged;
end;

procedure TWBNavig.Loaded;
var
  W, H: Integer;
begin
  inherited Loaded;
  W := Width;
  H := Height;
  AdjustSize (W, H);
  if (W <> Width) or (H <> Height) then
    inherited SetBounds (Left, Top, W, H);
  InitHints;
  ActiveChanged;
end;

{TWBNavButton}

destructor TWBNavButton.Destroy;
begin
  if FRepeatTimer <> nil then
    FRepeatTimer.Free;
  inherited Destroy;
end;

procedure TWBNavButton.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  inherited MouseDown (Button, Shift, X, Y);
  if nsAllowTimer in FNavStyle then
  begin
    if FRepeatTimer = nil then
      FRepeatTimer := TTimer.Create(Self);

    FRepeatTimer.OnTimer := TimerExpired;
    FRepeatTimer.Interval := InitRepeatPause;
    FRepeatTimer.Enabled  := True;
  end;
end;

procedure TWBNavButton.MouseUp(Button: TMouseButton; Shift: TShiftState;
                                  X, Y: Integer);
begin
  inherited MouseUp (Button, Shift, X, Y);
  if FRepeatTimer <> nil then
    FRepeatTimer.Enabled  := False;
end;

procedure TWBNavButton.TimerExpired(Sender: TObject);
begin
  FRepeatTimer.Interval := RepeatPause;
  if (FState = bsDown) and MouseCapture then
  begin
    try
      Click;
    except
      FRepeatTimer.Enabled := False;
      raise;
    end;
  end;
end;

procedure TWBNavButton.Paint;
var
  R: TRect;
begin
  inherited Paint;
  if (GetFocus = Parent.Handle) and
     (FIndex = TWBNavig (Parent).FocusedButton) then
  begin
    R := Bounds(0, 0, Width, Height);
    InflateRect(R, -3, -3);
    if FState = bsDown then
      OffsetRect(R, 1, 1);
    DrawFocusRect(Canvas.Handle, R);
  end;
end;

end.
